import json
import re
import aiohttp
from typing import Union
from ..utils.constant import Api, better_object
from random import shuffle, randint
from xmltodict import parse

Booru = Api()


class Paheal(object):
    """Paheal Client

    Methods
    -------
    search : function
        Search and gets images from paheal.

    search_image : function
        Search and gets images from paheal, but only returns image.

    """

    def __init__(self, api_key: str = "", user_id: str = ""):
        """Initializes paheal.

        Parameters
        ----------
        api_key : str
            Your API Key which is accessible within your account options page

        user_id : str
            Your user ID, which is accessible on the account options/profile page.
        """

        if api_key and user_id == "":
            self.api_key = None
            self.user_id = None
        else:
            self.api_key = api_key
            self.user_id = user_id

        self.specs = {"api_key": self.api_key, "user_id": self.user_id}

    async def search(
        self,
        query: str,
        limit: int = 100,
        page: int = 1,
        random: bool = True,
        gacha: bool = False,
    ) -> Union[aiohttp.ClientResponse, str]:

        """Search and gets images from paheal.

        Parameters
        ----------
        query : str
            The tags to search for.

        limit : int
            The limit of images to return.

        page : int
            The number of desired page

        random : bool
            Shuffle the whole dict, default is True.

        gacha : bool
            Get random single object, limit property will be ignored.

        Returns
        -------
        dict
            The json object returned by paheal.
        """
        if limit > 1000:
            raise ValueError(Booru.error_handling_limit)

        self.tags = query
        self.specs["tags"] = self.tags
        self.specs["limit"] = limit
        self.specs["page"] = page

        async with aiohttp.ClientSession() as session:
            async with session.get(Booru.paheal, params=self.specs) as resp:
                self.data = await resp.text()
                data_dict = parse(self.data)
                unsolved = json.dumps(data_dict)
                self.final = json.loads(unsolved)

                if "tag" not in self.final["posts"]:
                    raise ValueError(Booru.error_handling_null)

                self.not_random = self.final["posts"]["tag"]
                shuffle(self.not_random)

                try:
                    if gacha:
                        return better_object(
                            self.final["posts"]["tag"][
                                randint(0, len(self.final["posts"]["tag"]))
                            ]
                        )
                    elif random:
                        return better_object(self.final["posts"]["tag"])
                    else:
                        return better_object(self.not_random)

                except Exception as e:
                    raise Exception(f"Failed to get data: {e}")

    async def search_image(
        self, query: str, limit: int = 100, page: int = 1
    ) -> Union[aiohttp.ClientResponse, str]:

        """Gets images, meant just image urls from paheal.

        Parameters
        ----------
        query : str
            The tags to search for.

        limit : int
            The limit of images to return.

        page : int
            The number of desired page

        Returns
        -------
        list
            The list of image urls.

        """

        if limit > 1000:
            raise ValueError(Booru.error_handling_limit)

        self.tags = query
        self.specs["tags"] = self.tags
        self.specs["limit"] = limit
        self.specs["page"] = page

        try:
            async with aiohttp.ClientSession() as session:
                async with session.get(Booru.paheal, params=self.specs) as resp:
                    self.data = await resp.text()
                    data_dict = parse(self.data)
                    unsolved = json.dumps(data_dict)
                    self.final = json.loads(unsolved)

                    abc_kontol = self.final["posts"]["tag"]
                    ## extract all image urls
                    self.image_urls = []
                    for i in abc_kontol:  # paheal emang ngentot
                        self.image_urls.append(i["@file_url"])

                    shuffle(self.image_urls)
                    return better_object(self.image_urls)

        except Exception as e:
            raise Exception(f"Failed to get data: {e}")



import re
import aiohttp
from bs4 import BeautifulSoup
from typing import Union
from random import randint
from .constant import Api, ascii_to_str

Booru = Api()


async def request(
    site: str, params_x: dict, ua: dict = Booru.headers, block: str = ""
) -> Union[aiohttp.ClientResponse, list, None]:
    """Fetch the site

    Parameters
    ----------
    site : str
        The site to request
    params_x : dict
        The parameters to be passed
    ua : dict
        The user agent to be passed
    block : str
        The tags to be blocked

    Returns
    -------
    Union[aiohttp.ClientResponse, list, None]
        The response
    """

    if site == Booru.behoimi:
        ua = Booru.behoimi_bypass
    async with aiohttp.ClientSession(headers=ua) as session:
        async with session.get(site, params=params_x) as resp:
            data = await resp.json(content_type=None)
            if not data:
                raise Exception(Booru.error_handling_null)

            if "post" not in data:
                pattern = data

            elif "post" in data:
                pattern = data["post"]

            elif "images" in data:
                pattern = data["images"]

            try:
                for i in range(len(pattern)):
                    pattern[i]["tags"] = pattern[i]["tags"].split(" ")

                pattern = [i for i in pattern if not any(j in block for j in i["tags"])]

                return pattern

            except Exception as e:  ## danbooru
                if e.args[0] == "tags":
                    for i in range(len(pattern)):
                        pattern[i]["tag_string"] = pattern[i]["tag_string"].split(" ")

                    pattern = [
                        i
                        for i in pattern
                        if not any(j in block for j in i["tag_string"])
                    ]

                    return pattern

                else:  ## furry stuff sigh
                    return pattern


def roll(objek: list) -> dict:
    """Roll the object

    Parameters
    ----------
    objek : list
        The object to be rolled

    Returns
    -------
    dict
        The rolled object
    """

    return objek[randint(0, len(objek) - 1)]


async def request_wildcard(site: str, query: str) -> Union[list, None]:
    """Fetch wildcard

    Parameters
    ----------
    site : str
        The site to request
    query : str
        The query to request

    Returns
    -------
    Union[list, None]
        The response
    """

    if (
        site == Booru.gelbooru_wildcard
        or site == Booru.hypnohub_wildcard
        or site == Booru.rule34_wildcard
        or site == Booru.realbooru_wildcard
        or site == Booru.safebooru_wildcard
        or site == Booru.tbib_wildcard
        or site == Booru.xbooru_wildcard
    ):
        sorting = Booru.base_gelbooru_sorting_tags
        stop_disini_anjing = "index.php?page=post&s=list&tags="

    elif site == Booru.danbooru_wildcard or site == Booru.atfbooru_wildcard:
        sorting = Booru.base_danbooru_sorting_tags
        stop_disini_anjing = "/posts?tags="

    elif (
        site == Booru.yandere_wildcard
        or site == Booru.konachan_wildcard
        or site == Booru.konachan_net_wildcard
        or site == Booru.lolibooru_wildcard
    ):
        sorting = Booru.base_yandere_sorting_tags
        stop_disini_anjing = "/post?tags="

    async with aiohttp.ClientSession() as session:
        async with session.get(f"{site}{query}*{sorting}") as resp:
            soup = BeautifulSoup(await resp.text(), "html.parser")
            get_all = soup.find_all("a")
            all_tags = [tag["href"] for tag in get_all]

            all_tags = [tag for tag in all_tags if tag.startswith(stop_disini_anjing)]
            ## findall /posts?tags=(.*)
            validate_tags = [
                re.findall(r"tags=(.*)", tag)[0]
                for tag in all_tags
                if re.findall(r"tags=(.*)", tag)
            ]

            tags = [
                tag
                for tag in validate_tags
                if tag != "all"
                and tag != "order%3Arandom"
                and tag != "translation_request"
            ]

            return ascii_to_str(tags)



import aiohttp
from typing import Union
from ..utils.constant import Api, better_object, get_hostname
from random import shuffle, randint

Booru = Api()


class Furbooru(object):
    """furbooru wrapper

    Methods
    -------
    search : function
        Search and gets images from furbooru.

    search_image : function
        Gets images, image urls only from furbooru.

    """

    @staticmethod
    def append_object(raw_object: dict):
        """Extends new object to the raw dict

        Parameters
        ----------
        raw_object : dict
            The raw object returned by furbooru.

        Returns
        -------
        str
            The new value of the raw object
        """
        for i in range(len(raw_object)):
            if "id" in raw_object[i]:
                raw_object[i][
                    "post_url"
                ] = f"{get_hostname(Booru.furbooru)}/images/{raw_object[i]['id']}"

        return raw_object

    def __init__(self, key: str = ""):
        """Initializes furbooru.

        Parameters
        ----------
        key : str
            An optional authentication token. If omitted, no user will be authenticated.
        """

        if key:
            self.key = None
        else:
            self.key = key

        self.specs = {"key": self.key}

    async def search(
        self,
        query: str,
        limit: int = 100,
        page: int = 1,
        random: bool = True,
        gacha: bool = False,
    ):

        """Search and gets images from furbooru.

        Parameters
        ----------
        query : str
            The query to search for.


        limit : int
            The limit of images to return.

        page : int
            The number of desired page

        random : bool
            Shuffle the whole dict, default is True.

        gacha : bool
            Get random single object, limit property will be ignored.

        Returns
        -------
        dict
            The json object returned by furbooru.
        """

        if limit > 1000:
            raise ValueError(Booru.error_handling_limit)

        self.query = query
        self.specs["q"] = self.query
        self.specs["per_page"] = limit
        self.specs["page"] = page

        async with aiohttp.ClientSession() as session:
            async with session.get(Booru.furbooru, params=self.specs) as resp:
                self.data = await resp.json(content_type=None)
                self.final = self.data

                if not self.final["images"]:
                    raise ValueError(Booru.error_handling_null)

                self.not_random = Furbooru.append_object(self.final["images"])
                shuffle(self.not_random)

                try:
                    if gacha:
                        return better_object(
                            self.not_random[randint(0, len(self.not_random))]
                        )
                    elif random:
                        return better_object(self.not_random)
                    else:
                        return better_object(
                            Furbooru.append_object(self.final["images"])
                        )

                except Exception as e:
                    raise Exception(f"Failed to get data: {e}")

    async def search_image(self, query: str, limit: int = 100, page: int = 1):

        """Gets images, meant just image urls from furbooru.

        Parameters
        ----------
        query : str
            The query to search for.

        limit : int
            The limit of images to return.

        page : int
            The number of desired page

        Returns
        -------
        dict
            The json object returned by furbooru.

        """

        if limit > 1000:
            raise ValueError(Booru.error_handling_limit)

        self.query = query
        self.specs["q"] = self.query
        self.specs["per_page"] = limit
        self.specs["page"] = page

        try:

            async with aiohttp.ClientSession() as session:
                async with session.get(Booru.furbooru, params=self.specs) as resp:
                    self.data = await resp.json(content_type=None)
                    self.final = self.data

                    self.not_random = [
                        i["representations"]["full"] for i in self.final["images"]
                    ]
                    shuffle(self.not_random)
                    return better_object(self.not_random)

        except Exception as e:
            raise Exception(f"Failed to get data: {e}")



import json
import re
from booru import __version__


class Api:
    """Api class

    This class is used to parse the data from the api.

    Attributes:
        gelbooru (str): The base url for gelbooru.
        rule34 (str): The base url for rule34.
        tbib (str): The base url for tbib.
        safebooru (str): The base url for safebooru.
        xbooru (str): The base url for xbooru.
        realbooru (str): The base url for realbooru.
        hypnohub (str): The base url for hypnohub.
        danbooru (str): The base url for danbooru.
        atfbooru (str): The base url for atfbooru.
        yandere (str): The base url for yandere.
        konachan (str): The base url for konachan.
        konachan_net (str): The base url for konachan.net.
        lolibooru (str): The base url for lolibooru.
        e621 (str): The base url for e621.
        e926 (str): The base url for e926.
        derpibooru (str): The base url for derpibooru.
        furbooru (str): The base url for furbooru.
        behoimi (str): The base url for behoimi.
        paheal (str): The base url for paheal.
        e_handling_limit (str): The error message for the limit.
        e_handling_sameval (str): The error message for the same values.
        e_handling_cantparse (str): The error message for the parsing.
        e_handling_null (str): The error message for the null.
        e_handling_invalid_auth (str): The error message for the invalid auth.
        headers (dict): The headers request.
        bypass_headers (dict): The bypass headers request.

        base_gelbooru_sorting_tags (str): Pattern for gelbooru sorting tags. 
        base_danbooru_sorting_tags (str): Pattern for danbooru sorting tags.
        base_yande_sorting_tags (str): Pattern for yande sorting tags.

        gelbooru_wildcard (str): The base wildcard url for gelbooru search.
        hypnohub_wildcard (str): The base wildcard url for hypnohub search.
        rule34_wildcard (str): The base wildcard url for rule34 search.
        realbooru_wildcard (str): The base wildcard url for realbooru search.
        safebooru_wildcard (str): The base wildcard url for safebooru search.
        tbib_wildcard (str): The base wildcard url for tbib search.
        xbooru_wildcard (str): The base wildcard url for xbooru search.

        danbooru_wildcard (str): The base wildcard url for danbooru search.
        atfbooru_wildcard (str): The base wildcard url for atfbooru search.

        yandere_wildcard (str): The base wildcard url for yandere search.
        lolibooru_wildcard (str): The base wildcard url for lolibooru search.
        kona_wildcard (str): The base wildcard url for konachan search.
        konachan_net_wildcard (str): The base wildcard url for konachan.net search.
    """

    def __init__(
        self,
        BASE_headers={
            "User-Agent": f"booru/v{__version__} (https://pypi.org/project/booru);",
            "From": "hey@sinkaroid.org",
        },
        BYPASS_headers={
            "User-Agent": "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 6.1)",
            "Referer": "http://behoimi.org/data/ff/f3/",
            "From": "hey@sinkaroid.org",
        },
    ):

        self.gelbooru = "https://gelbooru.com/index.php?page=dapi&s=post&q=index"
        self.rule34 = "https://rule34.xxx/index.php?page=dapi&s=post&q=index"
        self.tbib = "https://tbib.org/index.php?page=dapi&s=post&q=index"
        self.safebooru = "https://safebooru.org/index.php?page=dapi&s=post&q=index"
        self.xbooru = "https://xbooru.com/index.php?page=dapi&s=post&q=index"
        self.realbooru = "https://realbooru.com/index.php?page=dapi&s=post&q=index"
        self.hypnohub = "https://hypnohub.net/index.php?page=dapi&s=post&q=index"
        self.danbooru = "https://danbooru.donmai.us/posts.json"
        self.atfbooru = "https://booru.allthefallen.moe/posts.json"
        self.yandere = "https://yande.re/post.json"
        self.konachan = "https://konachan.com/post.json"
        self.konachan_net = "https://konachan.net/post.json"
        self.lolibooru = "https://lolibooru.moe/post.json"
        self.e621 = "https://e621.net/posts.json"
        self.e926 = "https://e926.net/posts.json"
        self.derpibooru = "https://derpibooru.org/api/v1/json/search/images"
        self.furbooru = "https://furbooru.com/api/v1/json/search/images"
        self.behoimi = "http://behoimi.org/post/index.json"
        self.paheal = "https://rule34.paheal.net/api/danbooru/find_posts/index.xml"
        self.error_handling_limit = "there is a hard limit of 1000 posts per request."
        self.error_handling_sameval = "block values should not be hit to the query"
        self.error_handling_cantparse = "failed to get data, the api is misleading"
        self.error_handling_null = "no results, make sure you spelled everything right"
        self.error_handling_invalid_auth = "invalid api key or login"
        self.headers = BASE_headers
        self.behoimi_bypass = BYPASS_headers

        self.base_gelbooru_sorting_tags = "&sort=desc&order_by=index_count"
        self.base_danbooru_sorting_tags = "&search%5Border%5D=count"
        self.base_yandere_sorting_tags = "&type=&order=count"

        self.gelbooru_wildcard = "https://gelbooru.com/index.php?page=tags&s=list&tags="
        self.hypnohub_wildcard = "https://hypnohub.net/index.php?page=tags&s=list&tags="
        self.rule34_wildcard = "https://rule34.xxx/index.php?page=tags&s=list&tags="
        self.realbooru_wildcard = "https://realbooru.com/index.php?page=tags&s=list&tags="
        self.safebooru_wildcard = "https://safebooru.org/index.php?page=tags&s=list&tags="
        self.tbib_wildcard = "https://tbib.org/index.php?page=tags&s=list&tags="
        self.xbooru_wildcard = "https://xbooru.com/index.php?page=tags&s=list&tags="

        self.danbooru_wildcard = "https://danbooru.donmai.us/tags?commit=Search&search%5Bhide_empty%5D=yes&search%5Bname_or_alias_matches%5D="
        self.atfbooru_wildcard = "https://booru.allthefallen.moe/tags?commit=Search&search%5Bhide_empty%5D=yes&search%5Bname_or_alias_matches%5D="

        self.yandere_wildcard = "https://yande.re/tag?name="
        self.konachan_wildcard = "https://konachan.com/tag?name="
        self.konachan_net_wildcard = "https://konachan.net/tag?name="
        self.lolibooru_wildcard = "https://lolibooru.moe/tag?name="
        
BASE_URL = Api()


def list_api():
    """Returns the api url.

    Returns
    -------
    list
    """
    # create this list for mocking
    api_list = [
        BASE_URL.gelbooru,
        BASE_URL.rule34,
        BASE_URL.tbib,
        BASE_URL.safebooru,
        BASE_URL.xbooru,
        BASE_URL.realbooru,
        BASE_URL.hypnohub,
        BASE_URL.danbooru,
        BASE_URL.atfbooru,
        BASE_URL.yandere,
        BASE_URL.konachan,
        BASE_URL.konachan_net,
        BASE_URL.lolibooru,
        BASE_URL.e621,
        BASE_URL.e926,
        BASE_URL.derpibooru,
        BASE_URL.furbooru,
        BASE_URL.behoimi,
        BASE_URL.paheal,
    ]
    return api_list


def better_object(parser: dict):
    """Converts the json object to a more readable object.

    Parameters
    ----------
    parser : dict

    Returns
    -------
    str
        The new dictionaries with neat keys.

    """
    return json.dumps(parser, sort_keys=True, indent=4, ensure_ascii=False)


def parse_image(raw_object: dict):
    """Extracts the image url from the json object.
    Parameters
    ----------
    obj : dict
        The object to be parsed.
    Returns
    -------
    list
        The list of image urls.
    """
    if "post" not in raw_object:
        data = raw_object

    elif "post" in raw_object:
        data = raw_object["post"]

    try:
        images = [i["file_url"] for i in data]

    except:
        images = [i["file"]["url"] for i in data]  # furry stuff sigh

    images = list(dict.fromkeys(images))
    return images


def get_hostname(url: str):
    """Extract single hostname from the nested url

    Parameters
    ----------
    url : str

    Returns
    -------
    str
        The site contains protocol and hostname
    """
    return re.sub(r"(.*://)?([^/?]+).*", "\g<1>\g<2>", url)


def resolve(b_object: dict) -> dict:
    """Resolves the json object.

    Parameters
    ----------
    b_object : dict

    Returns
    -------
    dict
        raw json object
    """
    return json.loads(b_object)


def parse_image_danbooru(raw_object: dict) -> list:
    """Smh, it's danbooru though

    Parameters
    ----------
    raw_object : dict
        The object to be parsed.

    Returns
    -------
    list
        The list of image urls.
    """

    image = []

    try:
        for i in raw_object:
            image.append(i["file_url"])
        return image
    except KeyError:
        for i in raw_object:
            try:
                image.append(i["file_url"])
            except KeyError:
                pass
        return list(set(image))

def ascii_to_str(data: list):
    """ASCII to string 

    Parameters
    ----------
    data : list
        The data to be replaced

    Returns
    -------
    list
        The replaced data
    """

    bad_char = ["%28", "%29", "%2A", "%2F", "%3D", "%3A", "%27"]
    excpected_char = ["(", ")", "*", "/", "=", ":", "'"]
    for i in range(len(data)):
        for j in range(len(bad_char)):
            data[i] = data[i].replace(bad_char[j], excpected_char[j])

    return data




__version__ = "1.0.18"
from .client.gelbooru import Gelbooru
from .client.rule34 import Rule34
from .client.tbib import Tbib
from .client.safebooru import Safebooru
from .client.xbooru import Xbooru
from .client.realbooru import Realbooru
from .client.hypnohub import Hypnohub
from .client.danbooru import Danbooru
from .client.atfbooru import Atfbooru
from .client.yandere import Yandere
from .client.konachan import Konachan
from .client.konachan_net import Konachan_Net
from .client.lolibooru import Lolibooru
from .client.e621 import E621
from .client.e926 import E926
from .client.derpibooru import Derpibooru
from .client.furbooru import Furbooru
from .client.behoimi import Behoimi
from .client.paheal import Paheal
from .utils.constant import resolve



from typing import Union
from ..utils.fetch import request, roll
from ..utils.constant import Api, better_object, parse_image, get_hostname
from random import shuffle

Booru = Api()


class E926(object):
    """E926 wrapper

    Methods
    -------
    search : function
        Search and gets images from e926.

    search_image : function
        Search and gets images from e926, but only returns image.

    """

    @staticmethod
    def append_object(raw_object: dict):
        """Extends new object to the raw dict

        Parameters
        ----------
        raw_object : dict
            The raw object returned by e926.

        Returns
        -------
        str
            The new value of the raw object
        """
        for i in range(len(raw_object)):
            if "id" in raw_object[i]:
                raw_object[i][
                    "post_url"
                ] = f"{get_hostname(Booru.e926)}/posts/{raw_object[i]['id']}"

        return raw_object

    def __init__(self, api_key: str = "", user_id: str = ""):
        """Initializes e926.

        Parameters
        ----------
        api_key : str
            Your API Key (If possible)

        user_id : str
            Your user ID (If possible)
        """

        if api_key and user_id == "":
            self.api_key = None
            self.user_id = None
        else:
            self.api_key = api_key
            self.user_id = user_id

        self.specs = {"api_key": self.api_key, "user_id": self.user_id}

    async def search(
        self,
        query: str,
        limit: int = 100,
        page: int = 1,
        random: bool = True,
        gacha: bool = False,
    ) -> Union[list, str, None]:

        """Search and gets images from e926.

        Parameters
        ----------
        query : str
            The query to search for.

        limit : int
            The limit of images to return.

        page : int
            The number of desired page

        random : bool
            Shuffle the whole dict, default is True.

        gacha : bool
            Get random single object, limit property will be ignored.

        Returns
        -------
        dict
            The json object returned by e926.
        """

        if limit > 1000:
            raise ValueError(Booru.error_handling_limit)

        self.query = query
        self.specs["tags"] = self.query
        self.specs["limit"] = limit
        self.specs["page"] = page

        raw_data = await request(site=Booru.e926, params_x=self.specs, block="")
        self.appended = E926.append_object(raw_data["posts"])

        try:
            if gacha:
                return better_object(roll(self.appended))
            elif random:
                shuffle(self.appended)
                return better_object(self.appended)
            else:
                return better_object(E926.append_object(self.appended))
        except Exception as e:
            raise Exception(f"Failed to get data: {e}")

    async def search_image(
        self, query: str, limit: int = 100, page: int = 1
    ) -> Union[list, str, None]:

        """Gets images, meant just image urls from e926.

        Parameters
        ----------
        query : str
            The query to search for.

        limit : int
            The limit of images to return.

        page : int
            The number of desired page

        Returns
        -------
        list
            The list of image urls.

        """

        if limit > 1000:
            raise ValueError(Booru.error_handling_limit)

        self.query = query
        self.specs["tags"] = self.query
        self.specs["limit"] = limit
        self.specs["pid"] = page
        self.specs["json"] = "1"

        raw_data = await request(site=Booru.e926, params_x=self.specs, block="")
        self.appended = E926.append_object(raw_data["posts"])

        try:
            return better_object(parse_image(self.appended))
        except Exception as e:
            raise Exception(f"Failed to get data: {e}")



import re
from setuptools import setup

version = ""
with open("booru/__init__.py") as f:
    version = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE
    ).group(1)


requirements = []
with open("requirements.txt") as f:
    requirements = f.read().splitlines()


if not version:
    raise RuntimeError("version is not set")

readme = ""
with open("README.md", encoding="utf8") as f:
    readme = f.read()

setup(
    name="booru",
    author="sinkaroid",
    author_email="anakmancasan@gmail.com",
    version=version,
    long_description=readme,
    long_description_content_type = "text/markdown",
    url="https://github.com/sinkaroid/booru",
    project_urls={
        "CI": "https://github.com/sinkaroid/booru/actions",
        "Funding": "https://github.com/sponsors/sinkaroid",
        "Issue tracker": "https://github.com/sinkaroid/booru/issues/new/choose",
        "Documentation": "https://sinkaroid.github.io/booru",
        "Discord": "https://discord.gg/8wj4vM5hHM",
    },
    packages=["booru", "booru.client", "booru.utils"],
    license="MIT",
    classifiers=[
        "Framework :: AsyncIO",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Artistic Software",
        "Topic :: Games/Entertainment",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Build Tools",
    ],
    description="Python bindings for Booru imageboards",
    include_package_data=True,
    keywords=[
        "booru",
        "gelbooru",
        "rule34",
        "safebooru",
        "xbooru",
        "tbib",
        "realbooru",
        "hypnohub",
        "danbooru",
        "atfbooru",
        "yandere",
        "konachan",
        "konachan.net",
        "lolibooru",
        "e621",
        "e926",
        "derpibooru",
        "furbooru",
        "behoimi",
        "paheal",
    ],
    install_requires=requirements,
)



import re
from typing import Union
from ..utils.fetch import request, request_wildcard, roll
from ..utils.constant import Api, better_object, parse_image, get_hostname
from random import shuffle

Booru = Api()


class Gelbooru(object):
    """Gelbooru Client

    Methods
    -------
    search : function
        Search and gets images from gelbooru.

    search_image : function
        Search and gets images from gelbooru, but only returns image.

    find_tags : function
        Get the proper tags from gelbooru.

    """

    @staticmethod
    def append_object(raw_object: dict):
        """Extends new object to the raw dict

        Parameters
        ----------
        raw_object : dict
            The raw object returned by gelbooru.

        Returns
        -------
        str
            The new value of the raw object
        """
        for i in range(len(raw_object)):
            if "id" in raw_object[i]:
                raw_object[i][
                    "post_url"
                ] = f"{get_hostname(Booru.gelbooru)}/index.php?page=post&s=view&id={raw_object[i]['id']}"

        return raw_object

    def __init__(self, api_key: str = "", user_id: str = ""):
        """Initializes gelbooru.

        Parameters
        ----------
        api_key : str
            Your API Key (If possible)

        user_id : str
            Your user ID (If possible)
        """

        if api_key and user_id == "":
            self.api_key = None
            self.user_id = None
        else:
            self.api_key = api_key
            self.user_id = user_id

        self.specs = {"api_key": self.api_key, "user_id": self.user_id}

    async def search(
        self,
        query: str,
        block: str = "",
        limit: int = 100,
        page: int = 1,
        random: bool = True,
        gacha: bool = False,
    ) -> Union[list, str, None]:

        """Search method

        Parameters
        ----------
        query : str
            The query to search for.
        block : str
            The tags you want to block, separated by space.
        limit : int
            Expected number which is from pages
        page : int
            Expected number of page.
        random : bool
            Shuffle the whole dict, default is True.
        gacha : bool
            Get random single object, limit property will be ignored.

        Returns
        -------
        dict
            The json object (as string, you may need booru.resolve())
        """

        if limit > 1000:
            raise ValueError(Booru.error_handling_limit)

        if block and re.findall(block, query):
            raise ValueError(Booru.error_handling_sameval)

        self.query = query
        self.specs["tags"] = self.query
        self.specs["limit"] = limit
        self.specs["pid"] = page
        self.specs["json"] = "1"

        raw_data = await request(site=Booru.gelbooru, params_x=self.specs, block=block)
        self.appended = Gelbooru.append_object(raw_data)

        try:
            if gacha:
                return better_object(roll(self.appended))
            elif random:
                shuffle(self.appended)
                return better_object(self.appended)
            else:
                return better_object(Gelbooru.append_object(self.appended))
        except Exception as e:
            raise Exception(f"Failed to get data: {e}")

    async def search_image(
        self, query: str, block: str = "", limit: int = 100, page: int = 1
    ) -> Union[list, str, None]:

        """Parses image only

        Parameters
        ----------
        query : str
            The query to search for.
        block : str
            The tags you want to block, separated by space.
        limit : int
            Expected number which is from pages
        page : int
            Expected number of page.

        Returns
        -------
        dict
            The json object (as string, you may need booru.resolve())

        """

        if limit > 1000:
            raise ValueError(Booru.error_handling_limit)
        elif block and re.findall(block, query):
            raise ValueError(Booru.error_handling_sameval)

        self.query = query
        self.specs["tags"] = self.query
        self.specs["limit"] = limit
        self.specs["pid"] = page
        self.specs["json"] = "1"

        raw_data = await request(site=Booru.gelbooru, params_x=self.specs, block=block)
        self.appended = Gelbooru.append_object(raw_data)

        try:
            return better_object(parse_image(self.appended))
        except Exception as e:
            raise Exception(f"Failed to get data: {e}")

    async def find_tags(site: str, query: str) -> Union[list, str, None]:
        """Find tags

        Parameters
        ----------
        site : str
            The site to search for.
        query : str
            The tag to search for.

        Returns
        -------
        list
            The list of tags (as string, you may need booru.resolve())
        """
        try:
            data = await request_wildcard(site=Booru.gelbooru_wildcard, query=query)
            return better_object(data)

        except Exception as e:
            raise Exception(f"Failed to get data: {e}")
